#include "SphereProject.h"
#include "AdaptStationary.h"

/// RHS function
class F : public AbstractFunction<double, WorldVector<double> >
{
public:
  F(int degree) : AbstractFunction<double, WorldVector<double> >(degree) {}

  /// Implementation of AbstractFunction::operator().
  double operator()(const WorldVector<double>& x) const 
  {
    return -2.0 * x[0];
  }
};

Spheredemo::Spheredemo():
  sphere("sphere"),
  ballCenter(NULL),
  adaptInfo(NULL),
  adapt(NULL),
  matrixOperator(NULL),
  rhsOperator(NULL)
{}

Spheredemo::~Spheredemo() {
  if(matrixOperator != NULL)
    delete matrixOperator;
  if(rhsOperator != NULL)
    delete rhsOperator;
  if(ballCenter != NULL)
    delete ballCenter;
  if(adapt != NULL)
    delete adapt;
  if(adaptInfo != NULL)
    delete adaptInfo;
}

void Spheredemo::create(string& filename) {
   // ===== init parameters =====
  Parameters::init(false, filename);

  // ===== create projection =====
  ballCenter = new WorldVector< double >;
  ballCenter->set(0.0);
  new BallProject(1, BOUNDARY_PROJECTION, *ballCenter, 1.0);

  // ===== create and init the scalar problem ===== 
  sphere.initialize(INIT_ALL);

  // === create adapt info ===
  adaptInfo = new AdaptInfo("sphere->adapt");

  // === create adapt ===
  adapt = new AdaptStationary("sphere->adapt",
			       &sphere,
			       adaptInfo);
  
  // ===== create matrix operator =====
  matrixOperator = new Operator(sphere.getFeSpace());
  matrixOperator->addSecondOrderTerm(new Laplace_SOT);
  sphere.addMatrixOperator(matrixOperator);

  // ===== create rhs operator =====
  rhsOperator = new Operator(sphere.getFeSpace());

  int degree = sphere.getFeSpace()->getBasisFcts()->getDegree();

  rhsOperator->addZeroOrderTerm(new CoordsAtQP_ZOT(new F(degree)));
  sphere.addVectorOperator(rhsOperator);

}

int Spheredemo::solve(SolutionInformation& info) {
  int retCode = adapt->adapt();
  info.dofVec = sphere.getSolution();
  return retCode;
}
